/*
 * This program was inspired by the plugin binner.java
 * The idea behind is to reduce a stack in 3d. In addition,
 * the "average" could reduce the noise of the image. The negative effects of "average" were
 * not evaluated so far. It could be that "median" is the better choise.
 *
 * In 2d an alternative would be the function "Image -> Scale" of ImageJ
 * An alternative to the skip feature of this program would be "Image -> Stack -> Reduce Size"
 *
 * TODO: add median
 * TODO: add float stack
 *
 *
 * This code is in the public domain and may be freely used,
 * modified, or redistributed without permission.
 *
 * @author Karl-Heinz Kunzelmann <karl-heinz@kunzelmann.de>
 * @version Last Modified 16 August 2009
 */

//Achtung: mit diesem File können nur INT Ergebnisse generiert werden
// dh. Grey8 = Byte und Short funktioniert problemlos. Aber nicht float!
//Man müsste den Average mit double berechnen...
// Stand: 15.8.2009
// Median entfernt, Average und skip jetzt in 3d

import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.image.*;
import ij.plugin.filter.*;
import ij.plugin.*;
import ij.measure.*;


public class StackReducer_3d  implements PlugInFilter {
    ImagePlus imp;
    static int xshrink=2, yshrink=2, zshrink=2;
	private static String[] methods=new String[2];
	private static int methodindex = 0;
    int product;
    int pixel;
    int sum;
   


    public int setup(String arg, ImagePlus imp) {
        this.imp = imp;
        if (IJ.versionLessThan("1.28e")){
            return DONE;
        }
        else {
            return DOES_8G+DOES_16+NO_UNDO+STACK_REQUIRED;
        }
    }

    public void run(ImageProcessor ip) {
        if (showDialog()){
            shrink(ip);
        }
    }

    public void shrink(ImageProcessor ip) {
	
       
        int w = ip.getWidth()/xshrink; // TODO noch offen: genau nachdenken, was bei ungeraden Breiten, Höhen, Schichten-in-z passiert
        int h = ip.getHeight()/yshrink;
        int d = imp.getStackSize()/zshrink; //d = depth
        System.out.println("No. of slices in stack: "+d);

        	
        ColorModel cm=imp.createLut().getColorModel();
        ImageStack stack=imp.getStack();
	ImageStack newStack = new ImageStack (w,h,cm);


                
        // Initialize the ZProjector object.
        ZProjector zproj = new ZProjector(imp);

        // Important note: Slices are numbered 1,2,...,n where n is the
        // number of slices in the stack.

        for (int z=1; z<=imp.getStackSize(); z=z+zshrink) {
            IJ.showProgress(z/imp.getStackSize());

            // 
            // aus Running_ZProjector.java (Charlie Holly <caholly@colby.edu>) übernommen
            // Nutzt ZProjector.jave aus ImageJ für Projektion
            // Man legt nur den Anfangs und Endslice fest und die Projektionsmethode

            zproj.setStartSlice(z);
            zproj.setStopSlice(z+zshrink-1);
            ImagePlus projection;
            ImageProcessor theSlice = null;

            if (methodindex==0){
                zproj.setMethod(0);
            
            
                zproj.doProjection();
                // zproj.computeProjection(defaultMethod);
                projection =  zproj.getProjection();
                theSlice = projection.getProcessor(); // in dieser Version von theSlice sind die Slices schon gemittelt
            }
            else {
                theSlice=stack.getProcessor(z);  //war originalzeile
            }
                ImageProcessor theNewSlice=theSlice.createProcessor(w,h);
            if (theSlice.isInvertedLut()){
                theSlice.invert();
            }
            for (int y=0; y<h; y++){
                for (int x=0; x<w; x++) {
                        if (methodindex==0){         //0=average
                                 theNewSlice.putPixel(x, y, getAverage(theSlice, x, y));
                        }
                        else {

                                theNewSlice.putPixel(x, y, getSkip(theSlice, x, y));
                        }
                }
            }
                if (ip.isInvertedLut()) {
                    theSlice.invert();
                }
            newStack.addSlice("",theNewSlice);
        }
            String title="Reduced "+imp.getShortTitle();
            ImagePlus impNewImp = imp.createImagePlus();
            impNewImp.setStack(title, newStack);
            impNewImp.show();

            // set pixel distance for new ImagePlus

            Calibration calTheStack = imp.getCalibration();
            Calibration calNewStack = impNewImp.getCalibration();
		if (calTheStack.scaled()) {
			calNewStack.pixelWidth *= xshrink;
			calNewStack.pixelHeight *= yshrink;
                        calNewStack.pixelDepth *= zshrink;
		}
       
    }

    int getAverage(ImageProcessor ip, int x, int y) {
         //System.out.println("arrived in getAverage");
         sum = 0;       
         for (int y2=0; y2<yshrink; y2++) {
            for (int x2=0;  x2<xshrink; x2++) {
                pixel = ip.getPixel(x*xshrink+x2, y*yshrink+y2); 
                sum += pixel;
                if (x == 100){
                //System.out.println("Pixel = "+pixel+" :: Sum = "+sum);
                }
             }
        }
        
       sum = (int)((sum/product)+0.5);   //product = xshrink*yshrink;
                                         // Anmerkung: hier ist bisher keine Pruefung auf overflow erfolgt
       return sum;
    }

   
  int getSkip(ImageProcessor ip, int x, int y) {

         for (int y2=0; y2<yshrink; y2++) {
            for (int x2=0;  x2<xshrink; x2++) {

                pixel = ip.getPixel(x*xshrink+x2, y*yshrink+y2);


             }
        }

       return pixel;
    }

   /**
    * Builds a dialog to query users for reduction factor in each direction
    * In addition the user can select the method of the reduction
    * Average = writes the average of the pixels which will be removed
    *           into the newly calculated pixel
    * Skip = the supernummerary pixels are just deleted (=skipped)
    *
    *
    * 
    * @return boolean indicating that the GenericDialog was answered by the user
    */


    boolean showDialog() {
        GenericDialog gd = new GenericDialog("Stack Reducing in 3D");
        gd.addNumericField("X Shrink Factor:", xshrink, 0);
        gd.addNumericField("Y Shrink Factor:", yshrink, 0);
        gd.addNumericField("Z Shrink Factor:", zshrink, 0);
	methods[0]="average (in x,y AND z)"; 
	methods[1]="skip (in x,y AND z)";
        //methods[2]="";

	if (methodindex > methods.length){
		methodindex=0;
        }
	gd.addChoice ("Bin Method: ", methods, methods[methodindex]);
        gd.showDialog();
        if (gd.wasCanceled()) {
            return false;
        }
        xshrink = (int) gd.getNextNumber();
        yshrink = (int) gd.getNextNumber();
        zshrink = (int) gd.getNextNumber();
	methodindex=gd.getNextChoiceIndex();
        product = xshrink*yshrink;
        return true;
    }


}
